#include <windows.h>
#include <stdio.h>
#include <math.h>
#include <fcntl.h>
#include "concalc.h"

//moves up 1 row
void moverowup (void)
{
  if (currowb == currow)
    displaycell (curcol, currow, NOHIGHLIGHT, NOUPDATE);
  if (currow > toprow && currowb > toprow)
  {
    if (currowb == currow && curcolb == curcol)
    {
      currow--;
      currowb = currow;
    }
  }
  else
    if (toprow != 0)
    {
      scroll (DOWN, 1, LEFTMARGIN + 1, 3, 80, 
        SCREENROWS + 2, TEXTCOLOR);
      displayrow (--toprow, NOUPDATE);
      if (currowb == currow && curcolb == curcol)
      {
        currow--;
        currowb = currow;
      }
      setbottomrow ();
    }
}

//moves down one row
void moverowdown (void)
{
  if (currowb == currow)
    displaycell (curcol, currow, NOHIGHLIGHT, NOUPDATE);
  if (currow < bottomrow && currowb < bottomrow)
  {
    if (currowb == currow && curcolb == curcol)
    {
      currow++;
      currowb = currow;
    }
  }
  else 
    if (bottomrow < MAXROWS - 1)
    {
      scroll (UP, 1, LEFTMARGIN + 1, 3, 80, 
        SCREENROWS + 2, TEXTCOLOR);
      toprow++;
      if (currowb == currow && curcolb == curcol)
      {
        currow++;
        currowb = currow;
      }
      setbottomrow ();
      displayrow (bottomrow, NOUPDATE);
    }
}

//moves left one column
void movecolleft (void)
{
  int col, oldleftcol;
  unsigned char oldcolstart [SCREENCOLS];
  oldleftcol = leftcol;
  movmem (colstart, oldcolstart, sizeof (colstart));
  if (curcolb == curcol)
    displaycell (curcol, currow, NOHIGHLIGHT, NOUPDATE);
  if (curcol > leftcol && curcolb > leftcol)
  { 
    if (curcolb == curcol && currowb == currow)
    {
      curcol--;
      curcolb = curcol;
    }
  }
  else
    if (leftcol != 0)
    {
      if (currowb == currow && curcolb == curcol)
      {
        curcol--;
        curcolb = curcol;
      }
      leftcol--;
      setrightcol ();
      setleftcol ();
      if (oldleftcol <= rightcol)
        scroll (RIGHT, colstart[oldleftcol - leftcol] - 
          LEFTMARGIN, LEFTMARGIN + 1, 3, 80, 
          SCREENROWS + 2, TEXTCOLOR);
      clearlastcol ();
      for (col = leftcol; col <= oldleftcol - 1; col++)
        displaycol (col, NOUPDATE);
    }
}

//moves right one column
void movecolright(void)
{
  int col, oldleftcol, oldrightcol;
  unsigned char oldcolstart [SCREENCOLS];
  movmem (colstart, oldcolstart, sizeof (colstart));
  oldleftcol = leftcol;
  oldrightcol = rightcol;
  if (curcolb == curcol)
    displaycell (curcol, currow, NOHIGHLIGHT, NOUPDATE);
  if (curcol < rightcol && curcolb < rightcol)
  {
    if (curcolb == curcol && currowb == currow)
    {
      curcol++;
      curcolb = curcol;
    }
  }
  else 
    if (rightcol < (MAXCOLS - 1))
    {
      if (currowb == currow && curcolb == curcol)
      {
        curcol++;
        curcolb = curcol;
      }
      rightcol++;
      setleftcol ();
      setrightcol ();
      if (oldrightcol >= leftcol)
        scroll (LEFT, 
          oldcolstart [leftcol - oldleftcol] - 
          LEFTMARGIN, LEFTMARGIN + 1, 3, 80, 
          SCREENROWS + 2, TEXTCOLOR);
      clearlastcol ();
      for (col = oldrightcol + 1; col <= rightcol; col++)
        displaycol (col, NOUPDATE);
    }
}

//recalculates all of the numbers in the speadsheet
void recalc (void)
{
  int col, row, dummy;
  for (col = 0; col <= lastcol; col++)
    for (row = 0; row <= lastrow; row++)
      if ((cell [col] [row] != NULL) && 
        (cell [col] [row]->attrib == FORMULA))
        cell [col] [row]->v.f.fvalue = 
          parse (cell [col] [row]->v.f.formula, &dummy);
  displayscreen (UPDATE);
}

//changes and prints the current AutoCalc value on the screen
void changeautocalc (int newmode)
{
  char s [15];
  if (!autocalc && newmode) recalc ();
  autocalc = newmode;
  if (autocalc)
    strcpy (s, MSGAUTOCALC);
  else
    s [0] = 0;
  writef (73, 1, MSGAUTOCALCCOLOR, strlen (MSGAUTOCALC), s);
}

//changes and prints the current formula display value on the screen
void changeformdisplay (int newmode)
{
  char s [15];
  formdisplay = newmode;
  if (formdisplay)
    strcpy (s, MSGFORMDISPLAY);
  else
    s [0] = 0;
  writef (65, 1, MSGFORMDISPLAYCOLOR, 
    strlen (MSGFORMDISPLAY), s);
}

//edits a selected cell
void editcell (CELLPTR ecell)
{
  char s [MAXINPUT + 1];
  if (ecell == NULL) return;
  switch (ecell->attrib)
  {
    case TEXTCELL: strcpy (s, ecell->v.text); break;
    
    case VALUE:
      if (ecell->v.value == HUGE_VAL)
        strcpy (s, "0");
     else
       sprintf (s, "%.*f", MAXPLACES, ecell->v.value);
    break;
    
    case FORMULA: strcpy (s, ecell->v.f.formula); break;
  }
  if (!editstring (s, "", MAXINPUT) || (s [0] == 0)) 
    return;
  act (s);
  changed = TRUE;
}

//clears the current spreadsheet
void clearsheet (void)
{
  int col, row;
  for (row = 0; row <= lastrow; row++)
    for (col = 0; col <= lastcol; col++)
      deletecell (col, row, NOUPDATE);
  initvars ();
  setrightcol ();
  setbottomrow ();
  curcolb = curcol;
  currowb = currow;
  displayscreen (NOUPDATE);
  printfreemem ();
  changed = FALSE;
}

//loads a new spreadsheet
void loadsheet (char *filename)
{
  int size, allocated, reallastcol = 0, reallastrow = 0, 
    file, i, j, n, ignorep, ignorec;
  char cstr [5];
  HANDLE hFile;
  BOOL bSuccess = FALSE;
  DWORD dwFileSize;
  LPSTR pszFileText;
  if (filename [0] == 0)
  {
    writeprompt (MSGFILENAME);
    if (!editstring (filename, "", MAXINPUT)) return;
  }
  for (i = 0; filename [i] != 0 && 
    filename [i] != '.'; i++)
    str [i] = filename [i];
  str [i] = 0;
  lstrcpy (filename, str);
  lstrcat (filename, ".csv");
  if (access (filename, 0))
  {
    errormsg (MSGNOEXIST);
    return;
  }
  hFile = CreateFile (filename, GENERIC_READ,
    FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, 0);
  if (hFile == INVALID_HANDLE_VALUE)
  {
    errormsg (MSGNOOPEN);
    return;
  }
  clearsheet ();
  writef (1, 25, PROMPTCOLOR, 79, MSGLOADING);
  gotoxy (strlen (MSGLOADING) + 1, 25);
  dwFileSize = GetFileSize(hFile, NULL);
  if (dwFileSize != 0xFFFFFFFF)
    pszFileText = (LPSTR)GlobalAlloc(GPTR, 
      dwFileSize + 1);
    if (pszFileText != NULL)
      if (ReadFile (hFile, pszFileText, dwFileSize, 
        &written, NULL))
      {
         pszFileText [dwFileSize] = 0;
         bSuccess = TRUE;
      }
  if (bSuccess == FALSE)
  {
    errormsg (MSGIOERROR);
    return;
  }
  if (!CloseHandle (hFile))
  {
    errormsg (MSGIOERROR);
    return;
  }
  CharToOem (pszFileText, pszFileText);
  curcol = currow = ignorep = ignorec = 0;
  lstrcpy (str, "");
  loading = 1;
  for (i = 0; pszFileText [i] != 0; i++)
  {
    for (j = 0; j <= 3; j++) 
      cstr [j] = *(pszFileText + i + j);
    cstr [4] = 0;
    if (lstrcmpi (cstr, "sum(") == 0)
    {
      i = i + 3;
      ignorep = 1;
    }
    else if (pszFileText [i] == '"')
    {
      if (ignorec == 0)
        ignorec = 1;
      else
        ignorec = 0;
    }
    else if (pszFileText [i] == 10)
    {
      if (lstrlen (str) != 0) act (str);
      lstrcpy (str, "");
      curcol = 0;
      currow++;
    }
    else if ((pszFileText [i] == ',' ||
      pszFileText [i] == ';') && ignorec == 0)
    {
      if (lstrlen (str) != 0) act (str);
      lstrcpy (str, "");
      curcol++;
    }
    else if (pszFileText [i] == ')' && ignorep == 1)
      ignorep = 0;
    else if (pszFileText [i] != 13 &&
      pszFileText [i] != '=')
    {
      cstr [1] = 0;
      cstr [0] = pszFileText [i];
      if (lstrlen (str) < MAXINPUT) lstrcat (str, cstr);
    }
  }
  loading = 0;
  displayscreen (NOUPDATE);
  writef (1, 25, TEXTCOLOR, strlen (MSGLOADING), "");
  gotoxy (1, 25);
  printfreemem ();
  GlobalFree (pszFileText);
  curcol = currow = curcolb = currowb = 0;
  setrightcol ();
  displayscreen (NOUPDATE);
  changed = FALSE;
}

//saves the current spreadsheet
void savesheet (void)
{
  char filename [MAXINPUT + 1], eof = 26, *fstr, 
    cstr [2];
  int size, col, row, overwrite, file, i, j, n, cl, rw;
  CELLPTR cellptr;
  HANDLE hFile;
  filename [0] = 0;
  writeprompt (MSGFILENAME);
  if (!editstring (filename, "", MAXINPUT)) return;
  for (i = 0; filename [i] != 0 && filename [i] != '.';
    i++)
    str [i] = filename [i];
  str [i] = 0;
  lstrcpy (filename, str);
  lstrcat (filename, ".csv");
  if (!access (filename, 0))
    if (!getyesno (&overwrite, MSGOVERWRITE) || 
      (overwrite == 'N')) 
      return;
  writef (1, 25, PROMPTCOLOR, 79, MSGSAVING);
  gotoxy (strlen (MSGSAVING) + 1, 25);
  hFile = CreateFile (filename, GENERIC_WRITE, 0, 0,
     CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
  if (hFile == INVALID_HANDLE_VALUE)
  {
    errormsg (MSGNOOPEN);
    return;
  }
  cstr [1] = 0;
  for (row = 0; row <= lastrow; row++)
  {
    for (col = 0; col <= lastcol; col++)
    {
      cstr [0] = '"';
      lstrcpy (str, cstr);
      if (cell [col] [row] != NULL)
      {
        cellptr = cell [col] [row];
        switch (cellptr->attrib)
        {
          case TEXTCELL: 
            lstrcat (str, cellptr->v.text);
          break;

          case VALUE: 
            sprintf (str + 1, "%f", cellptr->v.value);
          break;
          
          case FORMULA:
            lstrcat (str, "=");
            for (i = 0; cellptr->v.f.formula [i] != 0;
              i++)
            {
              j = i;
              fstr = cellptr->v.f.formula + j;
              if (formulastart (&fstr, &cl, &rw) && 
                *(fstr++) == ':' &&
                formulastart (&fstr, &cl, &rw))
              {
                lstrcat (str, "SUM(");
                j = i + fstr - cellptr->v.f.formula - j;
                for (n = i; n < j; n++)
                {
                  cstr [0] = cellptr->v.f.formula [n];
                  lstrcat (str, cstr);
                }
                lstrcat (str, ")");
                i = j - 1;
              }
              else
              {
                cstr [0] = cellptr->v.f.formula [i];
                lstrcat (str, cstr);
              }
            }
          break;
        }
      }
      cstr [0] = '"';
      lstrcat (str, cstr);
      if (col != lastcol) lstrcat (str, ",");
      OemToChar (str, str);
      if (!WriteFile (hFile, str, lstrlen (str), 
      &written, NULL))
      {
        errormsg (MSGIOERROR);
        CloseHandle (hFile);
        return;
      }
    }
    if (!WriteFile (hFile, "\n", 1, &written, NULL))
    {
      errormsg (MSGIOERROR);
      CloseHandle (hFile);
      return;
    }
  }
  if (!CloseHandle (hFile))
  {
    errormsg (MSGIOERROR);
    return;
  }
  writef (1, 25, TEXTCOLOR, strlen (MSGSAVING), "");
  gotoxy (1, 25);
  changed = FALSE;
}

//returns the number of rows to print
int pagerows (int row, int toppage, int border)
{
  int rows;
  rows = toppage ? 66 - TOPMARGIN : 66;
  if (border) rows--;
  if (row + rows - 1 > lastrow)
    return (lastrow - row + 1);
  else
    return (rows);
}

//returns the number of columns to print starting at col
int pagecols (int col, int border, int columns)
{
  int len = ((col == 0) && (border)) ? columns - 
    LEFTMARGIN : columns;
  int firstcol = col;
  while ((len > 0) && (col <= lastcol)) 
    len -= colwidth [col++];
  if (len < 0) col--;
  return (col - firstcol);
}

//prints a copy of the spreadsheet to file or to printer
void printsheet (void)
{
  char filename [MAXINPUT + 1], s [133], 
    colstr [MAXCOLWIDTH + 1];
  FILE *file;
  int columns, counter1, counter2, counter3, col = 0, 
    row, border, toppage, lcol, lrow, dummy, printed, 
    oldlastcol;
  filename [0] = 0;
  lstrcpy (filename, "~.prn");
  if ((file = fopen (filename, "wt")) == NULL)
  {
    errormsg (MSGNOOPEN);
    return;
  }
  oldlastcol = lastcol;
  for (counter1 = 0; counter1 <= lastrow; counter1++)
    for (counter2 = lastcol; counter2 < MAXCOLS; 
      counter2++)
      if (format [counter2] [counter1] >= OVERWRITE) 
        lastcol = counter2;
  if (!getyesno (&columns, MSGCOLUMNS)) return;
  columns = (columns == 'Y') ? 131 : 79;
  if (!getyesno (&border, MSGBORDER)) return;
  border = (border == 'Y');
  while (col <= lastcol)
  {
    row = 0;
    toppage = TRUE;
    lcol = pagecols (col, border, columns) + col;
    while (row <= lastrow)
    {
      lrow = pagerows (row, toppage, border) + row;
      printed = 0;
      if (toppage)
        for (counter1 = 0; counter1 < TOPMARGIN; 
          counter1++)
        {
          fprintf (file, "\n");
          printed++;
        }
      for (counter1 = row; counter1 < lrow; counter1++)
      {
        if (border && counter1 == row && toppage)
        {
          if (col == 0 && border)
            sprintf (s, "%*s", LEFTMARGIN, "");
          else
            s [0] = 0;
          for (counter3 = col; counter3 < lcol; 
            counter3++)
          {
            centercolstring (counter3, colstr);
            strcat (s, colstr);
          }
          OemToChar (s, s);
          fprintf (file, "%s\n", s);
          printed++;
        }
        if (col == 0 && border)
          sprintf (s, "%-*d", LEFTMARGIN, counter1 + 1);
        else
          s [0] = 0;
        for (counter2 = col; counter2 < lcol; counter2++)
          strcat (s, cellstring (counter2, counter1, 
            &dummy, FORMAT));
        OemToChar (s, s);
        fprintf (file, "%s\n", s);
        printed++;
      }
      row = lrow;
      toppage = FALSE;
//      if (printed < 66) fprintf (file, "%c", FORMFEED);
    }
    col = lcol;
  }
  fclose (file);
  lastcol = oldlastcol;
  ShellExecute (NULL, "open", "wordpad", "~.prn", 
    "", SW_SHOWDEFAULT);
}

//sets the new column width for a selected column
void setcolwidth (int col)
{
  int width, row;
  writeprompt (MSGCOLWIDTH);
  if (!getint (&width, MINCOLWIDTH, MAXCOLWIDTH)) return;
  colwidth [col] = width;
  setrightcol ();
  if (rightcol < col)
  {
    rightcol = col;
    setleftcol ();
    setrightcol ();
  }
  for (row = 0; row <= lastrow; row++)
  {
    if (cell [col] [row] != NULL && 
      cell [col] [row]->attrib == TEXTCELL)
      clearoflags (col + 1, row, NOUPDATE);
    else
      clearoflags (col, row, NOUPDATE);
    updateoflags (col, row, NOUPDATE);
  }
  displayscreen (NOUPDATE);
  changed = TRUE;
}

//moves to a selected cell
void gotocell ()
{
  writeprompt (MSGGOTO);
  if (!getcell (&curcol, &currow)) return;
  leftcol = curcol;
  toprow = currow;
  setbottomrow ();
  setrightcol ();
  setleftcol ();
  displayscreen (NOUPDATE);
}

//prompts the user for a selected format and range of cells
void formatcells (void)
{
  int col, row, col1, col2, row1, row2, temp, 
    newformat = 0;
  writeprompt (MSGCELL1);
  if (!getcell (&col1, &row1)) return;
  writeprompt (MSGCELL2);
  if (!getcell (&col2, &row2)) return;
  if (col1 != col2 && row1 != row2)
    errormsg (MSGDIFFCOLROW);
  else
  {
    if (col1 > col2) swap (&col1, &col2);
    if (row1 > row2) swap (&row1, &row2);
    if (!getyesno (&temp, MSGRIGHTJUST)) return;
    newformat += (temp == 'Y') * RJUSTIFY;
    if (!getyesno (&temp, MSGDOLLAR)) return;
    newformat += (temp == 'Y') * DOLLAR;
    if (!getyesno (&temp, MSGCOMMAS)) return;
    newformat += (temp == 'Y') * COMMAS;
    if (newformat & DOLLAR)
      newformat += 2;
    else
    {
      writeprompt (MSGPLACES);
        if (!getint (&temp, 0, MAXPLACES)) return;
      newformat += temp;
    }
    for (col = col1; col <= col2; col++)
      for (row = row1; row <= row2; row++)
      {
        format [col] [row] = (format [col] [row] & 
          OVERWRITE) | newformat;
        if (col >= leftcol && col <= rightcol &&
          row >= toprow && row <= bottomrow)
          displaycell (col, row, NOHIGHLIGHT, NOUPDATE);
     }
  }
  changed = TRUE;
}

//deletes a column
void deletecol (int col)
{
  int counter, row;
  for (counter = 0; counter <= lastrow; counter++)
    deletecell (col, counter, NOUPDATE);
  printfreemem ();
  if (col != MAXCOLS - 1)
  {
    movmem (&cell [col + 1] [0], &cell [col] [0], 
      MAXROWS * sizeof (CELLPTR) * (MAXCOLS - col - 1));
    movmem (&format [col + 1] [0], &format [col] [0], 
      MAXROWS * (MAXCOLS - col - 1));
    movmem (&colwidth [col + 1], &colwidth [col], 
      MAXCOLS - col - 1);
  }
  setmem (&cell [MAXCOLS - 1] [0], 
    MAXROWS * sizeof (CELLPTR), 0);
  setmem (&format [MAXCOLS - 1] [0], MAXROWS, 
    DEFAULTFORMAT);
  colwidth [MAXCOLS - 1] = DEFAULTWIDTH;
  if ((lastcol >= col) && (lastcol > 0)) lastcol--;
  setrightcol ();
  if (curcol > rightcol)
  {
    rightcol++;
    setleftcol ();
  }
  clearlastcol ();
  for (counter = 0; counter <= lastcol; counter++)
    for (row = 0; row <= lastrow; row++)
    {
      if (cell [counter] [row] != NULL &&
        cell [counter] [row]->attrib == FORMULA)
        fixformula (counter, row, COLDEL, col);
      updateoflags (col, row, NOUPDATE);
    }
  while (col <= rightcol) displaycol (col++, NOUPDATE);
  changed = TRUE;
  recalc ();
}

//inserts a column
void insertcol (int col)
{
  int counter, row;
  if (lastcol == MAXCOLS - 1)
  {
    for (counter = 0; counter <= lastrow; counter++)
      deletecell (lastcol, counter, NOUPDATE);
    printfreemem ();
  }
  if (col != MAXCOLS - 1)
  {
    movmem (&cell [col] [0], &cell [col + 1] [0], 
      MAXROWS * sizeof (CELLPTR) * (MAXCOLS - col - 1));
    movmem (&format [col] [0], &format [col + 1] [0], 
      MAXROWS * (MAXCOLS - col - 1));
    movmem (&colwidth [col], &colwidth [col + 1], 
      MAXCOLS - col - 1);
  }
  setmem (&cell [col] [0], MAXROWS * sizeof (CELLPTR), 0);
  setmem(&format [col] [0], MAXROWS, DEFAULTFORMAT);
  colwidth [col] = DEFAULTWIDTH;
  lastcol = MAXCOLS - 1;
  setlastcol ();
  setrightcol ();
  if (curcol > rightcol)
  {
    rightcol++;
    setleftcol ();
  }
  for (counter = 0; counter <= lastcol; counter++)
    for (row = 0; row <= lastrow; row++)
    {
      if (cell [counter] [row] != NULL &&
        cell [counter] [row]->attrib == FORMULA)
        fixformula (counter, row, COLADD, col);
      updateoflags (col, row, NOUPDATE);
    }
  while (col <= rightcol) displaycol (col++, NOUPDATE);
  changed = TRUE;
  recalc ();
}

//deletes a row
void deleterow (int row)
{
  int counter, rowc;
  for (counter = 0; counter <= lastcol; counter++)
    deletecell (counter, row, NOUPDATE);
  printfreemem ();
  if (row != MAXROWS - 1)
    for (counter = 0; counter < MAXCOLS; counter++)
    {
      movmem (&cell [counter] [row + 1], 
        &cell [counter] [row], 
        sizeof (CELLPTR) * (MAXROWS - row - 1));
      movmem (&format [counter] [row + 1], 
        &format [counter] [row], MAXROWS - row - 1);
    }
  else
    for (counter = 0; counter <= lastcol; counter++)
    {
      cell [counter] [MAXROWS - 1] = NULL;
      format [counter] [MAXROWS - 1] = DEFAULTFORMAT;
    }
  if (lastrow >= row && lastrow > 0) lastrow--;
  for (counter = 0; counter <= lastcol; counter++)
    for (rowc = 0; rowc <= lastrow; rowc++)
      if (cell [counter] [rowc] != NULL &&
        cell [counter] [rowc]->attrib == FORMULA)
        fixformula (counter, rowc, ROWDEL, row);
  while (row <= bottomrow) displayrow (row++, NOUPDATE);
  changed = TRUE;
  recalc ();
}

//inserts a row
void insertrow (int row)
{
  int counter, rowc;
  if (lastrow == MAXROWS - 1)
  {
    for (counter = 0; counter <= lastcol; counter++)
      deletecell (counter, lastrow, NOUPDATE);
    printfreemem ();
  }
  if (row != MAXROWS - 1)
    for (counter = 0; counter < MAXCOLS; counter++)
    {
      movmem (&cell [counter] [row], 
        &cell [counter] [row + 1],
        sizeof (CELLPTR) * (MAXROWS - row - 1));
      movmem(&format [counter] [row], 
        &format [counter] [row + 1], MAXROWS - row - 1);
    }
  for (counter = 0; counter < MAXCOLS; counter++)
  {
    cell [counter] [row] = NULL;
    format [counter] [row] = DEFAULTFORMAT;
  }
  lastrow = MAXROWS - 1;
  setlastrow ();
  for (counter = 0; counter <= lastcol; counter++)
    for (rowc = 0; rowc <= lastrow; rowc++)
      if (cell [counter] [rowc] != NULL &&
        cell [counter] [rowc]->attrib == FORMULA)
        fixformula (counter, rowc, ROWADD, row);
  while (row <= bottomrow) displayrow (row++, NOUPDATE);
  changed = TRUE;
  recalc ();
}

//executes the commands in the spreadsheet menu
void smenu (void)
{
  char filename [MAXINPUT + 1];
  filename [0] = 0;
  switch (getcommand (SMENU, SCOMMAND))
  {
    case 0:
      checkforsave ();
      loadsheet (filename);
    break;
    
    case 1: savesheet (); break;
    case 2: printsheet (); break;
    
    case 3:
      checkforsave ();
      clearsheet ();
    break;
    
    case 4:
      writef (1, 24, COMMANDCOLOR, 79, MSGABOUT1);
      writef (1, 25, COMMANDCOLOR, 79, MSGABOUT2);
      getkey ();
      writef (1, 24, TEXTCOLOR, 79, "");
      writef (1, 25, TEXTCOLOR, 79, "");
    break;
  }
}

//executes the commands in the column menu
void cmenu (void)
{
  switch (getcommand (CMENU, CCOMMAND))
  {
    case 0: insertcol (curcol); break;
    case 1: deletecol (curcol); break;
    case 2: setcolwidth (curcol); break;
  }
}

//executes the commands in the row menu
void rmenu (void)
{
  switch (getcommand (RMENU, RCOMMAND))
  {
    case 0: insertrow (currow); break;
    case 1: deleterow (currow); break;
  }
}

//executes the commands in the utility menu
void umenu (void)
{
  switch (getcommand (UMENU, UCOMMAND))
  {
    case 0: recalc (); break;
    
    case 1:
      changeformdisplay (!formdisplay);
      displayscreen (UPDATE);
    break;
  }
}

//executes the commands in the main menu
void mainmenu (void)
{
  switch (getcommand (MENU, COMMAND))
  {
    case 0: smenu (); break;
    case 1: formatcells (); break;
    
    case 2:
      deletecell (curcol, currow, UPDATE);
      printfreemem ();
      if (autocalc) recalc ();
    break;
    
    case 3:
      if (curcolb == curcol && currowb == currow) 
        gotocell (); 
    break;

    case 4: 
      if (curcolb == curcol && currowb == currow) 
        cmenu (); 
    break;
    
    case 5: 
      if (curcolb == curcol && currowb == currow) 
        rmenu (); 
    break;
    
    case 6: editcell (curcell); break;
    case 7: umenu (); break;
    case 8: changeautocalc (!autocalc);  break;
    
    case 9:
      checkforsave ();
      stop = TRUE;
    break;
  }
}

