#include <windows.h>
#include <stdio.h>
#include "concalc.h"

extern int normalcursor;

//prints string on console at selected location in color
void writef (int col, int row, int color, int width, 
  char *format, ...)
{
  va_list arg_ptr;
  char output [81];
  int len, i;
  va_start (arg_ptr, format);
  vsprintf (output, format, arg_ptr);
  output [width] = 0;
  if ((len = strlen (output)) < width)
    setmem (&output [len], width - len, ' ');
  for (i = 0; i <= lstrlen (output); i++)
    if (output [i] == (char) 228 || 
      output [i] == (char) 229)
      output [i] = 147;
  SetConsoleTextAttribute (so, color);
  gotoxy (col, row);
  WriteConsole (so, output, lstrlen (output), &written, 
    NULL);
}

//scrolls an area of the screen
void scroll (int direction, int lines, int x1, int y1, 
  int x2, int y2, int attrib)
{
  if (lines == 0)
    window (x1, y1, x2, y2);
  else
    switch (direction)
    {
      case UP:
        movetext (x1, y1 + lines, x2, y2, x1, y1);
        window (x1, y2 - lines + 1, x2, y2);
      break;
      
      case DOWN:
        movetext (x1, y1, x2, y2 - lines, x1, y1 + lines);
        window (x1, y1, x2, y1 + lines - 1);
      break;
      
      case LEFT:
        movetext (x1 + lines, y1, x2, y2, x1, y1);
        window (x2 - lines + 1, y1, x2, y2);
      break;
      
      case RIGHT:
        movetext (x1, y1, x2 - lines, y2, x1 + lines, y1);
        window (x1, y1, x1 + lines - 1, y2);
      break;
    }
  coord.X = R.Left - 1;
  for (coord.Y = R.Top - 1; coord.Y <= R.Bottom - 1; 
    coord.Y++)
    FillConsoleOutputAttribute 
      (so, attrib, R.Right - R.Left + 1, coord, &written);
  for (coord.Y = R.Top - 1; coord.Y <= R.Bottom - 1; 
    coord.Y++)
    FillConsoleOutputCharacter 
      (so, ' ', R.Right - R.Left + 1, coord, &written);
}

//sets the shape of the cursor
VOID setcursor (UINT shape)
{
  if (shape == nocursor)
    curinfo.bVisible = FALSE;
  else
    curinfo.bVisible = TRUE;
  if (shape == shortcursor) curinfo.dwSize = 20;    
  if (shape == tallcursor) curinfo.dwSize = 90;
  if (shape == normalcursor) curinfo.dwSize = 5;
  SetConsoleCursorInfo (so, &curinfo);
}

//changes cursor shape based on the current insert mode
void changecursor (int insmode)
{
  if (insmode)
    setcursor (shortcursor);
  else
    setcursor (tallcursor);
}

//prints the column headings
void printcol (void)
{
 int col;
 char colstr [MAXCOLWIDTH + 1];

 writef (1, 2, HEADERCOLOR, 3, "");
 scroll (UP, 0, 1, 2, 80, 2, HEADERCOLOR);
 for (col = leftcol; col <= rightcol; col++)
 {
  centercolstring (col, colstr);
  writef (colstart [col - leftcol] + 1, 2, HEADERCOLOR, 
    colwidth [col], colstr);
 }
}

//clears any data left in the last column
void clearlastcol (void)
{
  int col;
  if ((col = colstart [rightcol - leftcol] + 
    colwidth [rightcol]) < 80)
    scroll (UP, 0, col + 1, 3, 80, SCREENROWS + 2, 
      TEXTCOLOR);
}

//prints the row headings
void printrow (void)
{
  int row;
  for (row = 0; row < SCREENROWS; row++)
    writef (1, row + 3, HEADERCOLOR, LEFTMARGIN, "%-d", 
      row + toprow + 1);
}

//displays the contents of a cell
void displaycell (int col, int row, int highlighting, 
  int updating)
{
  int color;
  char *s;
  if (loading == 1) return;
  if ((updating) &&
    ((cell [col] [row] == NULL) || 
    (cell [col] [row]->attrib != FORMULA)))
    return;
  s = cellstring (col, row, &color, FORMAT);
  if (highlighting || (col >= curcol && col <= curcolb &&
    row >= currow && row <= currowb &&
    (curcol != curcolb || currow != currowb)))
  {
    if (color == ERRORCOLOR)
      color = HIGHLIGHTERRORCOLOR;
    else
      color = HIGHLIGHTCOLOR;
  }
  writef (colstart [col - leftcol] + 1, 
    row - toprow + 3, color, colwidth [col], "%s", s);
}

//displays a column on the screen
void displaycol (int col, int updating)
{
  int row;
  for (row = toprow; row <= bottomrow; row++)
    displaycell (col, row, NOHIGHLIGHT, updating);
}

//displays a row on the screen
void displayrow (int row, int updating)
{
  int col;
  for (col = leftcol; col <= rightcol; col++)
    displaycell (col, row, NOHIGHLIGHT, updating);
}

//displays the current screen of the spreadsheet
void displayscreen (int updating)
{
  int row;
  for (row = toprow; row <= bottomrow; row++)
    displayrow (row, updating);
  clearlastcol ();
}

//clears the input line
void clearinput (void)
{
  scroll (UP, 0, 1, 25, 80, 25, TEXTCOLOR);
  gotoxy (1, 25);
}

//prints the type of cell and what is in it
void showcelltype (void)
{
  char colstr [3], *s;
  int color;
  formdisplay = !formdisplay;
  s = cellstring (curcol, currow, &color, NOFORMAT);
  colstring (curcol, colstr);
  if (curcell == NULL)
    writef (1, 23, CELLTYPECOLOR, 80, "%s%d %s", colstr, 
      currow + 1, MSGEMPTY);
  else
    switch (curcell->attrib)
    {
      case TEXTCELL:
        writef (1, 23, CELLTYPECOLOR, 80, "%s%d %s", 
          colstr, currow + 1, MSGTEXT);
      break;
      
      case VALUE:
        writef (1, 23, CELLTYPECOLOR, 80, "%s%d %s", 
          colstr, currow + 1, MSGVALUE);
      break;
      
      case FORMULA:
        writef (1, 23, CELLTYPECOLOR, 80, "%s%d %s", 
          colstr, currow + 1, MSGFORMULA);
      break;
    }
  writef (1, 24, CELLCONTENTSCOLOR, 80, "%s", s);
  formdisplay = !formdisplay;
}

//displays the entire screen
void redrawscreen(void)
{
  setrightcol ();
  setbottomrow ();
  writef (1, 1, MSGMEMORYCOLOR, strlen(MSGMEMORY), 
    MSGMEMORY);
  writef (29, 1, PROMPTCOLOR, strlen(MSGCOMMAND), 
    MSGCOMMAND);
  changeautocalc (autocalc);
  changeformdisplay (formdisplay);
  printfreemem ();
  displayscreen (NOUPDATE);
}

