#include <windows.h>
#include <math.h>
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include "concalc.h"

struct TOKENREC
{
  char state;
  union
  {
    double value;
    struct c_t
    {
      int row, col;
    } c;
    char funcname[MAXFUNCNAMELEN + 1];
  } x;
};


struct TOKENREC stack [PARSERSTACKSIZE], curtoken,
  token1, token2;
int stacktop, tokentype, error, counter;
char *input, isformula;

//returns TRUE if the string is a legal function, 
//FALSE otherwise
int isfunc (char *s)
{
  int len = strlen (s);
  if (strncmp (s, input, len) == 0)
  {
    strncpy (curtoken.x.funcname, input, len);
    curtoken.x.funcname [len] = 0;
    input += len;
    return (TRUE);
  }
  return (FALSE);
}

//pushes a new token onto the stack
void push (struct TOKENREC *token)
{
  if (stacktop == PARSERSTACKSIZE - 1)
  {
    errormsg (MSGSTACKERROR);
    error = TRUE;
  }
  else
    stack [++stacktop] = *token;
}

//pops the top token off of the stack
struct TOKENREC pop (void)
{
  return (stack [stacktop--]);
}

//finds the value of a particular cell
double cellvalue (int col, int row)
{
  if (cell [col] [row] == NULL) return (0);
  if (cell [col] [row]->attrib == TEXTCELL) 
    return (HUGE_VAL);
  if (cell [col] [row]->attrib == FORMULA)
    return (cell [col] [row]->v.f.fvalue);
  return (cell [col] [row]->v.value);
}

//shifts a token onto the stack
void shift (int state)
{
  curtoken.state = state;
  push (&curtoken);
  tokentype = nexttoken ();
}

//gets the next token from the input stream
int nexttoken (void)
{
  char *start, numstring [80];
  int decimal, len, numlen;
  while (*input == ' ') input++;
  if (*input == 0) return (EOLN);
  if (strchr ("0123456789.", *input))
  {
    start = input;
    len = 0;
    decimal = FALSE;
    while (isdigit (*input) || (*input == '.' && 
      !decimal))
    {
      if (*input == '.') decimal = TRUE;
      input++;
      len++;
    }
    if (len == 1 && start [0] == '.') return (BAD);
    if (*input == 'E')
    {
      input++;
      len++;
      if (strchr ("+-", *input) != NULL)
      {
        input++;
        len++;
      }
      numlen = 0;
      while (isdigit (*input) && ++numlen <= 3)
      {
        input++;
        len++;
      }
    }
    strncpy (numstring, start, len);
    numstring [len] = 0;
    curtoken.x.value = atof (numstring);
    if (errno == ERANGE) return (BAD);
    return (NUM);
  }
  else if (isalpha (*input))
  {
    if (isfunc ("ABS") || isfunc ("ACOS") || 
      isfunc ("ASIN") || isfunc ("ATAN") || 
      isfunc ("COSH") || isfunc ("COS") ||
      isfunc ("EXP") || isfunc ("LOG10") || 
      isfunc ("LOG") || isfunc ("POW10") || 
      isfunc ("ROUND") || isfunc ("SINH") ||
      isfunc ("SIN") || isfunc ("SQRT") || 
      isfunc ("SQR") || isfunc ("TANH") || 
      isfunc ("TAN") || isfunc ("TRUNC")) 
      return (FUNC);
    if (formulastart (&input, &curtoken.x.c.col, 
      &curtoken.x.c.row))
    {
      isformula = TRUE;
      return (CELL);
    }
    else
      return (BAD);
  }
  else 
    switch (*(input++))
    {
      case '+': return (PLUS);
      case '-': return (MINUS);
      case '*': return (TIMES);
      case '/': return (DIVIDE);
      case '^': return (EXP);
      case ':': return (COLON);
      case '(': return (OPAREN);
      case ')': return (CPAREN);
      default: return (BAD);
  }
}

//parses the string s, returns the value of the 
//evaluated string, and puts the attribute in att: 
//TEXTCELL = 0, CONSTANT = 1, FORMULA = 2
double parse (char *s, int *att)
{
  char accepted = FALSE;
  char copy [80];
  error = FALSE;
  isformula = FALSE;
  input = copy;
  strupr (strcpy (copy, s));
  stacktop = -1;
  curtoken.x.value = 0;
  shift (SFIRST);
  do
  {
    switch (stack [stacktop].state)
    {
      case SFIRST: case SOPAREN: case SPLUS: case SMINUS: 
      case STIMES: case SDIVIDE: case SPOW: 
      case SFOPAREN:
        if (tokentype == NUM)
          shift (SNUM);
        else if (tokentype == CELL)
          shift (SCELL);
        else if (tokentype == FUNC)
          shift (SFUNC);
        else if (tokentype == MINUS)
          shift (SNEGATE);
        else if (tokentype == OPAREN)
          shift (SOPAREN);
        else
          error = TRUE;
      break;
      
      case SFIRSTPLUS:
        if (tokentype == EOLN)
          accepted = TRUE;
        else if (tokentype == PLUS)
          shift (SPLUS);
        else if (tokentype == MINUS)
          shift (SMINUS);
        else
          error = TRUE;
      break;
      
      case SFIRSTTIMES:
        if (tokentype == TIMES)
          shift (STIMES);
        else if (tokentype == DIVIDE)
          shift (SDIVIDE);
        else
          reduce (RFIRSTTIMES);
      break;
      
      case SFIRSTPOW: reduce (RFIRSTPOW); break;
      
      case SFIRSTNEGATE:
        if (tokentype == EXP)
          shift (SPOW);
        else
          reduce (RFIRSTNEGATE);
      break;
      
      case SNEGATE:
        if (tokentype == NUM)
          shift (SNUM);
        else if (tokentype == CELL)
          shift (SCELL);
        else if (tokentype == FUNC)
          shift (SFUNC);
        else if (tokentype == OPAREN)
          shift (SOPAREN);
        else
          error = TRUE;
      break;
      
      case SFIRSTSUM: reduce (RFIRSTSUM); break;
      
      case SCELL:
        if (tokentype == COLON)
          shift (SCOLON);
        else
          reduce (RCELLVALUE);
      break;
      
      case SFIRSTFUNC: reduce (RFIRSTFUNC); break;
      case SNUM: reduce (RNUM); break;
      
      case SFUNC:
        if (tokentype == OPAREN)
          shift (SFOPAREN);
        else
          error = TRUE;
      break;
      
      case SDONEGATE: reduce (RNEGATE); break;
      
      case SCOLON:
        if (tokentype == CELL)
          shift (SDOSUM);
        else
          error = TRUE;
      break;
      
      case SOPARENPLUS:
        if (tokentype == PLUS)
          shift (SPLUS);
        else if (tokentype == MINUS)
          shift (SMINUS);
        else if (tokentype == CPAREN)
          shift (SCPAREN);
        else
          error = TRUE;
      break;
      
      case SPLUSTIMES:
        if (tokentype == TIMES)
          shift (STIMES);
        else if (tokentype == DIVIDE)
          shift (SDIVIDE);
        else
          reduce (RPLUS);
      break;
      
      case SMINUSTIMES:
        if (tokentype == TIMES)
          shift (STIMES);
        else  if (tokentype == DIVIDE)
          shift (SDIVIDE);
        else
          reduce (RMINUS);
      break;
      
      case SDOTIMES: reduce (RTIMES); break;
      case SDODIVIDE: reduce (RDIVIDE); break;
      case SDOPOW: reduce (RPOW); break;
      case SDOSUM: reduce (RSUM); break;
      case SCPAREN: reduce (RCPAREN); break;
      
      case SFOPARENPLUS:
        if (tokentype == PLUS)
          shift (SPLUS);
        else if (tokentype == MINUS)
          shift (SMINUS);
        else if (tokentype == CPAREN)
          shift (SDOFUNC);
        else 
          error = TRUE;
      break;
      
      case SDOFUNC: reduce (RFUNC); break;
      case SERROR: error = TRUE; break;
    }
  }
  while ((!accepted) && (!error));
  if (error)
  {
    *att = TEXTCELL;
    return (0);
  }
  if (isformula)
    *att = FORMULA;
  else
    *att = VALUE;
  strcpy (s, copy);
  return (stack [stacktop].x.value);
}

//completes a reduction
void reduce (int reduction)
{
  switch (reduction)
  {
    case RPLUS:
      token1 = pop ();
      pop ();
      token2 = pop ();
      curtoken.x.value = token1.x.value + token2.x.value;
    break;
    
    case RMINUS:
      token1 = pop ();
      pop ();
      token2 = pop ();
      curtoken.x.value = token2.x.value - token1.x.value;
    break;
    
    case RTIMES:
      token1 = pop ();
      pop ();
      token2 = pop ();
      curtoken.x.value = token1.x.value * token2.x.value;
    break;
    
    case RDIVIDE:
      token1 = pop ();
      pop ();
      token2 = pop ();
      if (token1.x.value == 0)
        curtoken.x.value = HUGE_VAL;
      else
        curtoken.x.value = token2.x.value / 
          token1.x.value;
    break;
    
    case RPOW:
      token1 = pop ();
      pop ();
      token2 = pop ();
      curtoken.x.value = pow (token2.x.value, 
        token1.x.value);
    break;
    
    case RNEGATE:
      token1 = pop ();
      pop ();
      curtoken.x.value = -token1.x.value;
    break;
    
    case RSUM:
      token1 = pop ();
      pop ();
      token2 = pop ();
      curtoken.x.value = 0;
      if (token1.x.c.row == token2.x.c.row)
      {
        if (token1.x.c.col < token2.x.c.col)
          error = TRUE;
        else
          for (counter = token2.x.c.col; 
            counter <= token1.x.c.col; counter++)
            curtoken.x.value += cellvalue (counter, 
              token1.x.c.row);
      }
      else if (token1.x.c.col == token2.x.c.col)
      {
        if (token1.x.c.row < token2.x.c.row)
          error = TRUE;
        else
          for (counter = token2.x.c.row; 
            counter <= token1.x.c.row; counter++)
              curtoken.x.value += 
                cellvalue (token1.x.c.col, counter);
      }
      else
        error = TRUE;
    break;
    
    case RCELLVALUE:
      curtoken = pop ();
      curtoken.x.value = cellvalue (curtoken.x.c.col, 
        curtoken.x.c.row);
    break;
    
    case RCPAREN:
      pop ();
      curtoken = pop ();
      pop ();
    break;
    
    case RFUNC:
      pop ();
      curtoken = pop ();
      pop ();
      token1 = pop ();
      if (strcmp (token1.x.funcname, "ABS") == 0)
        curtoken.x.value = fabs (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "ACOS") == 0)
        curtoken.x.value = acos (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "ASIN") == 0)
        curtoken.x.value = asin (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "ATAN") == 0)
        curtoken.x.value = atan (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "COSH") == 0)
        curtoken.x.value = cosh (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "COS") == 0)
        curtoken.x.value = cos (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "EXP") == 0)
        curtoken.x.value = exp (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "LOG10") == 0)
        curtoken.x.value = log10 (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "LOG") == 0)
        curtoken.x.value = log (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "ROUND") == 0)
        curtoken.x.value = (int) (curtoken.x.value + 0.5);
      else if (strcmp (token1.x.funcname, "POW10") == 0)
        curtoken.x.value = pow (10, curtoken.x.value);
      else if (strcmp (token1.x.funcname, "SINH") == 0)
        curtoken.x.value = sinh (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "SIN") == 0)
        curtoken.x.value = sin (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "SQRT") == 0)
        curtoken.x.value = sqrt (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "SQR") == 0)
        curtoken.x.value *= curtoken.x.value;
      else if (strcmp (token1.x.funcname, "TANH") == 0)
        curtoken.x.value = tanh (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "TAN") == 0)
        curtoken.x.value = tan (curtoken.x.value);
      else if (strcmp (token1.x.funcname, "TRUNC") == 0)
        curtoken.x.value = (int) curtoken.x.value;
    break;
    
    case RFIRSTTIMES: case RFIRSTPOW: case RFIRSTNEGATE: 
    case RFIRSTSUM: case RFIRSTFUNC: case RNUM: 
      curtoken = pop (); 
    break;
  }

  if (reduction <= 3) //RPLUS, RMINUS, RFIRSTTIMES
    switch (stack [stacktop].state)
    {
      case SFIRST: curtoken.state = SFIRSTPLUS; break;
      case SOPAREN: curtoken.state = SOPARENPLUS; break;
      case SFOPAREN: curtoken.state = SFOPARENPLUS; break;
    }
  else if (reduction <= 6) //RTIMES, RDIVIDE, RFIRSTPOW
    switch (stack [stacktop].state)
    {
      case SFIRST: case SOPAREN: case SFOPAREN: 
        curtoken.state = SFIRSTTIMES;
      break;
      
      case SPLUS: curtoken.state = SPLUSTIMES; break;
      case SMINUS: curtoken.state = SMINUSTIMES; break;
    }
  else if (reduction <= 8) //RPOW, RFIRSTNEGATE
    switch (stack [stacktop].state)
    {
      case SFIRST: case SOPAREN: case SPLUS: case SMINUS: 
      case SFOPAREN: 
        curtoken.state = SFIRSTPOW;
      break;
      
      case STIMES: curtoken.state = SDOTIMES; break;
      case SDIVIDE: curtoken.state = SDODIVIDE; break;
      case SPOW: curtoken.state = SDOPOW; break;
    }
  else if (reduction <= 10) //RNEGATE, RFIRSTSUM
    switch (stack [stacktop].state)
    {
      case SFIRST: case SOPAREN: case SPLUS: case SMINUS: 
      case STIMES: case SDIVIDE: case SPOW: 
      case SFOPAREN: 
        curtoken.state = SFIRSTNEGATE;
      break;
    }
  else if (reduction <= 12) //RSUM, RFIRSTFUNC
    switch (stack [stacktop].state)
    {
      case SFIRST: case SOPAREN: case SPLUS: case SMINUS: 
      case STIMES: case SDIVIDE: case SPOW: 
      case SFOPAREN:
        curtoken.state = SFIRSTSUM;
      break;
      
      case SNEGATE: curtoken.state = SDONEGATE;
    }
  else //RCELLVALUE, RCPAREN, RNUM, RFUNC
    switch (stack [stacktop].state)
    {
      case SFIRST: case SNEGATE: case SOPAREN: case SPLUS: 
      case SMINUS: case STIMES: case SDIVIDE: 
      case SPOW: case SFOPAREN: 
        curtoken.state = SFIRSTFUNC;
      break;
      
      default: curtoken.state = SERROR; break;
    }
   
  push (&curtoken);
}

