#include <windows.h>
#include <math.h>
#include <malloc.h>
#include <ctype.h>
#include <stdio.h>

#include "concalc.h"

//allocates space for a text cell
int alloctext (int col, int row, char *s)
{
  int size;
  CELLPTR cellptr;

  cellptr = (CELLPTR) malloc (sizeof ( struct CELLREC ));
  cellptr->attrib = TEXTCELL;
  strcpy (cellptr->v.text, s);
  cell [col] [row] = cellptr;
  return (TRUE);
}

//allocates space for a value cell
int allocvalue (int col, int row, double amt)
{
  CELLPTR cellptr;
  cellptr = (CELLPTR) malloc (sizeof (struct CELLREC));
  cellptr->attrib = VALUE;
  cellptr->v.value = amt;
  cell [col] [row] = cellptr;
  return (TRUE);
}

//allocates space for a formula cell
int allocformula (int col, int row, char *s, double amt)
{
  int size;
  CELLPTR cellptr;
  cellptr = (CELLPTR) malloc (sizeof (struct CELLREC));
  cellptr->attrib = FORMULA;
  strcpy (cellptr->v.f.formula, s);
  cellptr->v.f.fvalue = amt;
  cell [col] [row] = cellptr;
  return (TRUE);
}

//deletes a cell
void deletecell (int col, int row, int display)
{
  CELLPTR cellptr = cell [col] [row];
  if (cellptr == NULL) return;
  switch (cellptr->attrib)
  {
    case TEXTCELL:
      clearoflags (col + 1, row, display);
    break;
   
    case VALUE: break;
    case FORMULA: break;
  }
  format [col] [row] &= ~OVERWRITE;
  free (cell [col] [row]);
  cell [col] [row] = NULL;
  if (col == lastcol) setlastcol ();
  if (row == lastrow) setlastrow ();
  updateoflags (col, row, display);
  changed = TRUE;
}

//prints the amount of free memory
void printfreemem (void)
{
  memstatus.dwLength = sizeof (MEMORYSTATUS);
  GlobalMemoryStatus (&memstatus);
  writef (strlen (MSGMEMORY) + 2, 1, MEMORYCOLOR, 10,
    "%lu", memstatus.dwAvailPhys);
}

//returns the width in spaces of row
int rowwidth (int row)
{
  return ((row == 0) ? 1 : (int) log10 (row + 1) + 1);
}

//returns TRUE if the string is the start of a formula, 
//FALSE otherwise, also returns the column and row of 
//the formula
int formulastart (char **input, int *col, int *row)
{
  int len, maxlen = rowwidth (MAXROWS);
  char *start, numstring [10];
  if (!isalpha (**input)) return (FALSE);
  *col = *((*input)++) - 'A';
  if (isalpha (**input))
  {
    *col *= 26;
    *col += *((*input)++) - 'A' + 26;
  }
  if (*col >= MAXCOLS) return (FALSE);
  start = *input;
  for (len = 0; len < maxlen; len++)
    if (!isdigit (*((*input)++)))
    {
      (*input)--;
      break;
    }
  if (len == 0) return (FALSE);
  strncpy (numstring, start, len);
  numstring [len] = 0;
  *row = atoi (numstring) - 1;
  if (*row >= MAXROWS || *row == -1) return (FALSE);
  return (TRUE);
}

//prints an error message at the bottom of the screen
void errormsg (char *s)
{
  writef (1, 25, ERRORCOLOR, 79, "%s  %s", s, 
    MSGKEYPRESS);
  gotoxy (strlen (s) + strlen (MSGKEYPRESS) + 3, 25);
  getkey ();
  gotoxy (1, 25);
  writef (1, 25, TEXTCOLOR, 79, "");
}

//modifies a formula when its column or row designations 
//need to change
void fixformula (int col, int row, int action, int place)
{
  char *colstart, *rowstart, s [6], 
    newformula [MAXINPUT + 1], *curpos = newformula;
  int fcol, frow;
  CELLPTR cellptr = cell [col] [row];
  double value;
  strcpy (newformula, cellptr->v.f.formula);
  while (*curpos != 0)
  {
    if (formulastart (&curpos, &fcol, &frow))
    {
      rowstart = curpos - rowwidth (frow);
      colstart = rowstart - ((fcol > 25) ? 2 : 1);
      switch (action)
      {
        case COLADD:
          if (fcol < place) break;
          if (fcol == 25)
          {
            if (strlen (newformula) == MAXINPUT)
            {
              deletecell (col, row, NOUPDATE);
              alloctext (col, row, newformula);
              return;
            }
            movmem (colstart, colstart + 1, 
              strlen (colstart) + 1);
          }
          colstring (fcol + 1, s);
          movmem (s, colstart, strlen (s));
        break;
        
        case ROWADD:
          if (frow < place) break;
          if (rowwidth (frow + 1) != rowwidth (frow))
          {
            if (strlen (newformula) == MAXINPUT)
            {
              deletecell (col, row, NOUPDATE);
              alloctext (col, row, newformula);
              return;
            }
            movmem (rowstart, rowstart + 1, 
              strlen (rowstart) + 1);
          }
          sprintf (s, "%d", frow + 2);
          movmem (s, rowstart, strlen (s));
        break;
        
        case COLDEL:
          if (fcol <= place) break;
          if (fcol == 26) 
            movmem (colstart + 1, colstart, 
              strlen (colstart) + 1);
          colstring (fcol - 1, s);
          movmem (s, colstart, strlen (s));
        break;
        
        case ROWDEL:
          if (frow <= place) break;
          if (rowwidth (frow) != rowwidth (frow - 1))
            movmem (rowstart + 1, rowstart, 
              strlen (rowstart) + 1);
          sprintf (s, "%d", frow);
          movmem (s, rowstart, strlen (s));
        break;
      }
    }
    else
      curpos++;
  }
  if (strlen (newformula) != 
    strlen (cellptr->v.f.formula))
  {
    value = cellptr->v.f.fvalue;
    deletecell (col, row, NOUPDATE);
    allocformula (col, row, newformula, value);
  }
  else
    strcpy (cellptr->v.f.formula, newformula);
}

//fix copied formula
void fixformula2 (int col, int row)
{
  char *colstart, *rowstart, s [6], 
    newformula [MAXINPUT + 1], *curpos = newformula;
  int fcol, frow, rowdif;
  CELLPTR cellptr = cell [col] [row];
  double value;
  strcpy (newformula, cellptr->v.f.formula);
  while (*curpos != 0)
  {
    if (formulastart (&curpos, &fcol, &frow))
    {
      rowstart = curpos - rowwidth (frow);
      colstart = rowstart - ((fcol > 25) ? 2 : 1);
      if (fcol >= curcolc && fcol <= curcolcb &&
        frow >= currowc && frow <= currowcb)
      {
        if (fcol <= 25 && fcol + curcol - curcolc > 25)
        {
          if (strlen (newformula) == MAXINPUT)
          {
            deletecell (col, row, NOUPDATE);
            alloctext (col, row, newformula);
            return;
          }
          movmem (colstart, colstart + 1, 
            strlen (colstart) + 1);
          curpos++;
          rowstart++;
        }
        if (fcol > 25 && fcol + curcol - curcolc <= 25)
        {
          movmem (colstart + 1, colstart, 
            strlen (colstart) + 1);
          curpos--;
          rowstart--;
        }

        rowdif = rowwidth (frow + currow - currowc) - 
          rowwidth (frow);
        if (rowdif > 0)
        {
          if (strlen (newformula) + rowdif > MAXINPUT)
          {
            deletecell (col, row, NOUPDATE);
            alloctext (col, row, newformula);
            return;
          }
          movmem (rowstart, rowstart + rowdif, 
            strlen (rowstart) + 1);
          curpos = curpos + rowdif;
        }
        if (rowdif < 0)
        {
          movmem (rowstart - rowdif, rowstart, 
            strlen (rowstart) + 1);
          curpos = curpos + rowdif;
        }
            
        colstring (fcol + curcol - curcolc, s);
        movmem (s, colstart, strlen (s));
        sprintf (s, "%d", frow + currow - currowc + 1);
        movmem (s, rowstart, strlen (s));
      }
    }
    else
      curpos++;
  }
  strcpy (cellptr->v.f.formula, newformula);
}

//changes a column number to a string
void colstring (int col, char *colstr)
{
  setmem (colstr, 3, 0);
  if (col < 26)
    colstr [0] = col + 'A';
  else
  {
    colstr [0] = (col / 26) - 1 + 'A';
    colstr [1] = (col % 26) + 'A';
  }
}

//changes a column to a centered string
void centercolstring (int col, char *colstr)
{
  char s [3];
  int spaces1, spaces2;
  colstring (col, s);
  spaces1 = (colwidth [col] - strlen (s)) >> 1;
  spaces2 = colwidth [col] - strlen (s) - spaces1;
  sprintf (colstr, "%*s%s%*s", spaces1, "", s, spaces2, 
    "");
}

//sets the value of leftcol based on the value of 
//rightcol
void setleftcol (void)
{
  int total = 80, col = 0;
  while (total >= LEFTMARGIN && rightcol - col >= 0)
  {
    colstart [SCREENCOLS - col - 1] = 
      total - colwidth [rightcol - col];
    total -= colwidth [rightcol - col++];
  }
  if (total >= LEFTMARGIN) col++;
  movmem (&colstart [SCREENCOLS - col + 1], 
    colstart, col - 1);
  leftcol = rightcol - col + 2;
  total = colstart [0] - LEFTMARGIN;
  if (total != 0)
    for (col = leftcol; col <= rightcol; col++)
      colstart [col - leftcol] -= total;
  printcol();
}

//sets the value of rightcol based on the value of 
//leftcol
void setrightcol (void)
{
  int total = LEFTMARGIN, col = 0;
  do
  {
    colstart [col] = total;
    total += colwidth [leftcol + col++];
  } 
  while (total <= 80 && leftcol + col <= MAXCOLS);
  rightcol = leftcol + col - 2;
  printcol ();
}

//figures out the value of toprow based on the value of bottomrow
void settoprow (void)
{
  if (bottomrow - SCREENROWS < -1) bottomrow = 19;
  toprow = bottomrow - 19;
  printrow ();
}

//figures out the value of bottomrow based on the value of toprow
void setbottomrow (void)
{
  if (toprow + SCREENROWS > MAXROWS) 
    toprow = MAXROWS - 20;
  bottomrow = toprow + 19;
  printrow ();
}

//sets the value of lastcol based on the current value
void setlastcol (void)
{
  register int row, col;
  for (col = lastcol; col >= 0; col--)
    for (row = 0; row <= lastrow; row++)
      if (cell [col] [row] != NULL)
      {
        lastcol = col;
        return;
      }
  lastcol = 0;
}

//sets the value of lastrow based on the current value
void setlastrow (void)
{
  register int row, col;
  for (row = lastrow; row >= 0; row--)
    for (col = 0; col <= lastcol; col++)
      if (cell [col] [row] != NULL)
      {
        lastrow = row;
        return;
      }
  lastrow = 0;
}

//acts on a particular input
void act (char *s)
{
  int attrib, allocated;
  double value;
  deletecell (curcol, currow, UPDATE);
  value = parse (s, &attrib);
  switch (attrib)
  {
    case TEXTCELL:
      allocated = alloctext (curcol, currow, s);
      if (allocated) 
        displaycell (curcol, currow, NOHIGHLIGHT, 
          NOUPDATE);
    break;
    
    case VALUE: 
      allocated = allocvalue (curcol, currow, value); 
    break;
    
    case FORMULA:
      allocated = allocformula (curcol, currow, s, value);
    break;
  }
  if (allocated)
  {
    format [curcol] [currow] &= ~OVERWRITE;
    clearoflags (curcol + 1, currow, UPDATE);
    if (attrib == TEXTCELL) 
      setoflags (curcol, currow, UPDATE);
    if (curcol > lastcol) lastcol = curcol;
    if (currow > lastrow) lastrow = currow;
    if (autocalc) recalc ();
  }
  else
    errormsg (MSGLOMEM);
  printfreemem ();
}

//acts on a content of the copied cell
void act2 (char *s, double value, int attrib)
{
  int allocated;
  changed = TRUE;
  deletecell (j, i, UPDATE);
  switch (attrib)
  {
    case TEXTCELL:
      allocated = alloctext (j, i, s);
    break;
    
    case VALUE: 
      allocated = allocvalue (j, i, value); 
    break;
    
    case FORMULA:
      allocated = allocformula (j, i, s, value);
    break;
  }
  if (allocated)
  {
    format [j] [i] &= ~OVERWRITE;
    clearoflags (j + 1, i, UPDATE);
    if (attrib == TEXTCELL) 
      setoflags (j, i, UPDATE);
    if (j > lastcol) lastcol = j;
    if (i > lastrow) lastrow = i;
    if (autocalc) recalc ();
  }
  else
    errormsg (MSGLOMEM);
  printfreemem ();
}

//sets the overwrite flag on cells starting at 
//(col + 1, row), returns the number of the column after
//the last column set
int setoflags (int col, int row, int display)
{
  int len;
  len = strlen (cell [col] [row]->v.text) - 
    colwidth [col];
  while (++col < MAXCOLS && len > 0 && 
    cell [col] [row] == NULL)
  {
    format [col] [row] |= OVERWRITE;
    len -= colwidth [col];
    if (display && col >= leftcol && col <= rightcol)
      displaycell (col, row, NOHIGHLIGHT, NOUPDATE);
  }
  return (col);
}

//clears the overwrite flag on cells starting at 
//(col, row)
void clearoflags (int col, int row, int display)
{
  while (format [col] [row] >= OVERWRITE && 
    col < MAXCOLS && cell [col] [row] == NULL)
  {
    format [col] [row] &= ~OVERWRITE;
    if (display && col >= leftcol &&  col <= rightcol)
      displaycell (col, row, NOHIGHLIGHT, NOUPDATE);
    col++;
  }
}

//starting in col, moves back to the last TEXTCELL and 
//updates all flags
void updateoflags (int col, int row, int display)
{
  while (cell [col] [row] == NULL && col-- > 0);
  if (cell [col] [row] != NULL && 
    cell [col] [row]->attrib == TEXTCELL && col >= 0)
    setoflags (col, row, display);
}

//sets the string representation of text
void textstring (char *instring, char *outstring, 
  int col, int fvalue, int formatting)
{
  char *just, *ljust = "%-*s", *rjust = "%*s";
  if ((fvalue & RJUSTIFY) && formatting)
    just = rjust;
  else
    just = ljust;
  sprintf (outstring, just, colwidth [col], instring);
  if (formatting) outstring [colwidth [col]] = 0;
}

//sets the string representation of a value
void valuestring (CELLPTR cellptr, double value, 
  char *vstring, int col, int fvalue, int *color, 
  int formatting)
{
  char s [81];
  char *fstring;
  int width, pos;
  if (value == HUGE_VAL)
  {
    strcpy (vstring, MSGERROR);
    *color = ERRORCOLOR;
  }
  else
  {
    if (formatting)
    {
      sprintf (vstring, "%1.*f", fvalue & 15, 
        cellptr->v.value);
      if (fvalue & COMMAS)
      {
        pos = strcspn (vstring, ".");
        while (pos > 3)
        {
          pos -= 3;
          if (vstring [pos - 1] != '-')
          {
            movmem (&vstring[pos], &vstring[pos + 1], 
              strlen(vstring) - pos + 1);
            vstring [pos] = ',';
          }
        }
      }
      if (fvalue & DOLLAR)
      {
        if (vstring [0] == '-')
        {
          fstring = " $";
          width = colwidth [col] - 2;
        }
        else
        {
          fstring = " $ ";
          width = colwidth [col] - 3;
        }
      }
      else
      {
        fstring = "";
        width = colwidth [col];
      }
      strcpy (s, vstring);
      if (fvalue & RJUSTIFY)
      {
        if (strlen (vstring) > width)
          vstring [width] = 0;
        else
          sprintf (vstring, "%*s", width, s);
      }
      else
        sprintf (vstring, "%-*s", width, s);
      movmem (vstring, &vstring [strlen (fstring)], strlen (vstring) + 1);
      strncpy(vstring, fstring, strlen (fstring));
    }
    else
      sprintf (vstring, "%.*f", MAXPLACES, value);
    *color = VALUECOLOR;
  }
}

//creates an output string for the data in the cell in 
//(col, row), and also returns the color of the cell
char *cellstring (int col, int row, int *color, 
  int formatting)
{
  CELLPTR cellptr = cell [col] [row];
  int newcol, formatvalue;
  static char s [81], temp [MAXCOLWIDTH + 1];
  char *p;
  double value;
  if (cellptr == NULL)
  {
    if (!formatting || format [col] [row] < OVERWRITE)
    {
      sprintf (s, "%*s", colwidth [col], "");
      *color = BLANKCOLOR;
    }
    else
    {
      newcol = col;
      while (cell [--newcol] [row] == NULL);
      p = cell [newcol] [row]->v.text;
      while (newcol < col) p += colwidth [newcol++];
      strncpy (temp, p, colwidth [col]);
      temp [colwidth [col]] = 0;
      sprintf (s, "%s%*s", temp, colwidth [col] - 
        strlen (temp), "");
      *color = TEXTCOLOR;
    }
  }
  else
  {
    formatvalue = format [col] [row];
    switch (cellptr->attrib)
    {
      case TEXTCELL:
        textstring (cellptr->v.text, s, col, 
          formatvalue, formatting);
        *color = TEXTCOLOR;
      break;
      
      case FORMULA:
        if (formdisplay)
        {
          textstring (cellptr->v.f.formula, s, col, 
            formatvalue, formatting);
          *color = FORMULACOLOR;
          break;
        }
        else
          value = cellptr->v.f.fvalue;
          
      case VALUE :
        if (cellptr->attrib == VALUE) 
          value = cellptr->v.value;
          valuestring (cellptr, value, s, col, 
            formatvalue, color, formatting);
      break;
    }
  }
  return (s);
}

//prints a prompt on the screen
void writeprompt (char *prompt)
{
  writef (1, 24, PROMPTCOLOR, 80, prompt);
}

//swaps the first and second values
void swap (int *val1, int *val2)
{
  int temp;
  temp = *val1;
  *val1 = *val2;
  *val2 = temp;
}

//if the spreadsheet has been changed, will ask the user 
//if they want to save it
void checkforsave (void)
{
  int save;
  if (changed && getyesno (&save, MSGSAVESHEET) && 
    save == 'Y')
    savesheet ();
}

//initializes various global variables
void initvars(void)
{
  leftcol = toprow = curcol = currow = lastcol = 
    lastrow = curcolb = currowb = currowc = curcolc = 
    currowcb = curcolcb = 0;
  setmem (colwidth, sizeof (colwidth), DEFAULTWIDTH);
  setmem (cell, sizeof (cell), 0);
  setmem (format, sizeof (format), DEFAULTFORMAT);
}

//reads in a command and acts on it
int getcommand (char *msgstr, char *comstr)
{
  int ch, counter, len = strlen (msgstr);
  scroll (UP, 0, 1, 24, 80, 24, TEXTCOLOR);
  for (counter = 0; counter < len; counter++)
  {
    if (isupper (msgstr [counter]))
      writef (counter + 1, 24, COMMANDCOLOR, 1, "%c", 
        msgstr [counter]);
    else
      writef (counter + 1, 24, LOWCOMMANDCOLOR, 1, "%c",
        msgstr [counter]);
  }
  do
    ch = toupper (getkey ());
  while (strchr (comstr, ch) == NULL && ch != ESC);
  clearinput ();
  if (ch == ESC)
    return (-1);
  else
    return 
      (strlen (comstr) - strlen (strchr (comstr, ch)));
}

//fills memory
int setmem (void *Destination, unsigned int Length, 
  char Fill)
{
  FillMemory (Destination, Length, Fill);
}

//moves memory
int movmem (const void *Source, void *Destination, 
  unsigned int Length)
{
  MoveMemory (Destination, Source, Length); 
}

//clears screen
void clrscr (void)
{
  coord.X = 0;
  for (coord.Y = 0; coord.Y <= 24; coord.Y++)
    FillConsoleOutputCharacter 
      (so, ' ', 80, coord, &written);
  for (coord.Y = 0; coord.Y <= 79; coord.Y++)
    FillConsoleOutputAttribute 
      (so, TEXTCOLOR, 80, coord, &written);
}

//sets console cursor position
void gotoxy (int x, int y)
{
  SetConsoleCursorPosition (so, (COORD) {x - 1, y - 1});
}

//sets small rectangle to be used later
void window (int left, int top, int right, int bottom)
{
  R.Left = left;
  R.Top = top;
  R.Right = right;
  R.Bottom = bottom;
}

//moves a rectangular area of the screen
VOID movetext (INT left, INT top, INT right, INT bottom,
  INT newleft, INT newtop)
{
  R.Left = left - 1;
  R.Top = top - 1;
  R.Right = right - 1;
  R.Bottom = bottom - 1;
  ScrollConsoleScreenBuffer (so, &R, NULL,
    (COORD) {newleft - 1, newtop - 1}, &charinfo);
}

